#!/bin/sh
# FILE: mmGraph.tcl
#
# The control widget and network wrapping for an instance of the
# Ow_GraphWin class.
#
# This comment continues on in Tcl, but not in sh \
exec wish "$0" ${1+"$@"}

# App global variables:
# _mmgraph() general state information array
array set _mmgraph {
    saveall 1
    autosave 0
    autosave_handle {}
    autosave_column_index_list {}
    autosave_strfmt {}
    autosave_numfmt {}
}
# Autosave Note: All the autosave elts are setup inside the SaveFile proc,
#   including the autosave element itself.

#    _mmgpsc() display configuration array:
array set _mmgpsc {
    title  {}
    autolabel 1
    xlabel {}
    ylabel {}
    xdefaultlabel {}
    ydefaultlabel {}
    autolimits 1
    xmin   0
    xmax   1
    ymin   0
    ymax   1
    ptbufsize 0
    oldptbufsize 0
}
## Note: Default ptbufsize 0 => no limit

# This code can be source'd from inside another script by setting
# no_net and no_event_loop as desired before sourcing this file.
# If no_net is set true, use the proc "AddTriples { triples }" to add
# data.
if {![info exists no_net]} {
    set no_net 0          ;# Default is to setup server
}
if {![info exists no_event_loop]} {
    set no_event_loop 0   ;# Default is to enter the event loop
}

# Process command line arguments
foreach option $argv {
    if { [string match "--" $option] } { break }  ;# No more options
    if {[regexp -nocase -- {^--?no_?net$} $option]} {
	set no_net 1
    }
}

# Support libraries
package require Oc
package require Tk
package require Ow

Oc_Main SetAppName mmGraph
Oc_Main SetVersion 1.0.1.1
regexp \\\044Date:(.*)\\\044 {$Date: 1998/11/06 00:28:44 $} _ date
Oc_Main SetDate [string trim $date]
# regexp \\\044Author:(.*)\\\044 {$Author: dgp $} _ author
# Oc_Main SetAuthor [Oc_Person Lookup [string trim $author]]
Oc_Main SetAuthor [Oc_Person Lookup donahue]
Oc_Main SetHelpURL [Oc_Url FromFilename [file join [file dirname \
        [file dirname [file dirname [Oc_DirectPathname [info \
        script]]]]] doc userguide Data_Graph_Display_mmGraph.html]]


# Death
proc Die { win } {
    if {![string match "." $win]} { return }  ;# Child destroy event
    global protocol
    CloseAutoSave
    if {[info exists protocol]} {
	$protocol Delete; after 2000 exit
    } else {
	exit
    }
}
bind . <Destroy> {+Die %W}

# Save data routines
proc SaveRecord { record } {
    # Used to implement autosave functionality.  For better performance,
    # the SaveFile proc has a builtin version of this.
    # NOTE: This routine _assumes_ that all the _mmgraph autosave elts
    #       are valid.
    global _mmgraph
    set fileid $_mmgraph(autosave_handle)
    set column_index $_mmgraph(autosave_column_index_list)
    set strfmt $_mmgraph(autosave_strfmt)
    set numfmt $_mmgraph(autosave_numfmt)
    # Indent past heading   "# Columns:"
    puts -nonewline $fileid "          "
    foreach col $column_index {
	set value [lindex $record $col]
	if {[string match {} $value]} {
	    puts -nonewline $fileid [format $strfmt "      {}"]
	} else {
	    puts -nonewline $fileid [format $numfmt $value]
	}
    }
    puts $fileid {}   ;# newline
    flush $fileid
}

proc CloseAutoSave {} {
    # Close current autosave file, if any.  Always use this routine to
    # close a open autosave file.  This routine should be robust,
    # meaning it can be called without checking the state of
    # _mmgraph(autosave).
    global _mmgraph
    set fileid $_mmgraph(autosave_handle)
    if {![string match {} $fileid]} {
	set _mmgraph(autosave_handle) {}
	puts $fileid "# Table End"
	close $fileid
    }
    set _mmgraph(autosave) 0
}

proc SaveFile { filename setup_autosave } {
    global oommf _mmgpsc _mmgraph

    # Close any current autosave file
    CloseAutoSave

    # New file output
    set numfmt " %- 15.8g"  ;# For table body.
    set strfmt " %-15s"     ;# For column headers
    set oldfile [file exists $filename]
    if {[catch {open $filename "a+"} fileid]} {
	global oommf
	Ow_NonFatalError \
		"[Oc_Main GetAppName]: Unable to open $filename for appending"
	return
    }
    if {!$oldfile} { puts $fileid "# ODT 1.0" }
    puts $fileid "# Table Start"
    if {![string match {} $_mmgpsc(title)]} {
	puts $fileid "# Title: $_mmgpsc(title)"
    } else {
	puts $fileid "## [Oc_Main GetAppName] output"
    }
    global menu_select data_value data_index data_unit

    # Determing output table column labels
    if {$_mmgraph(saveall)} {
	# Save all data
	set column_index {}
	set column_list $data_index(all)
	foreach ordinate $column_list {
	    lappend column_index $data_index($ordinate)
	}
    } else {
	# Save data selected by displayed graphs
	set column_list {}
	set column_index {}
	if {[string match {} $menu_select(x)]} {
	    return ;# No abscissa selected
	}
	lappend column_list $menu_select(x)
	lappend column_index $data_index($menu_select(x))
	foreach ordinate $data_index(all) {
	    if {[info exists menu_select(y,$ordinate)] && \
		    $menu_select(y,$ordinate) } {
		# Label $ordinate is turned on in y-axis menu
		lappend column_list $ordinate
		lappend column_index $data_index($ordinate)
	    }
	}
    }

    # Build column headers
    set column_labels {}
    set unit_labels {}
    set unit_offsets {}
    foreach col $column_list {
	# Compress and strip newlines
	regsub -all "\[ \r\t\n\]" $col { } clab
	lappend column_labels $clab
	regsub -all "\[ \r\t\n\]" $data_unit($col) { } ulab
	lappend unit_labels $ulab
	set lendiff [expr [string length [list $clab]] \
		- [string length [list $ulab]]]
	if {$lendiff>1} {
	    lappend unit_offsets [expr $lendiff/2]
	} else {
	    lappend unit_offsets 0
	}
    }

    # Write column header
    puts -nonewline $fileid   "# Columns: "
    foreach lab $column_labels {
	puts -nonewline $fileid [format $strfmt [list $lab]]
    }
    puts -nonewline $fileid "\n# Units:   "
    foreach lab $unit_labels off $unit_offsets {
	puts -nonewline $fileid [format $strfmt	\
		[format "%*s%s" $off {} [list $lab]]]
    }
    puts $fileid {} ;# newline

    # Write out all rows of table
    for {set i $data_value(count,start)} {$i<$data_value(count,end)} {incr i} {
	# Indent past heading   "# Columns:"
        puts -nonewline $fileid "          "
	set record $data_value($i)
	foreach col $column_index {
	    set value [lindex $record $col]
	    if {[string match {} $value]} {
		puts -nonewline $fileid [format $strfmt "      {}"]
	    } else {
		puts -nonewline $fileid [format $numfmt $value]
	    }
	}
	puts $fileid {}   ;# newline
    }

    if { $setup_autosave } {
	flush $fileid
	set _mmgraph(autosave_handle) $fileid
	set _mmgraph(autosave_column_index_list) $column_index
	set _mmgraph(autosave_strfmt) $strfmt
	set _mmgraph(autosave_numfmt) $numfmt
	set _mmgraph(autosave) 1
    } else {
	set _mmgraph(autosave) 0
	puts $fileid "# Table End"
	close $fileid
    }
}

proc SaveFileOptionBoxSetup { widget frame } {
    global _mmgraph
    set _mmgraph(SFOB_autosave) $_mmgraph(autosave)
    checkbutton $frame.autosave -text "Auto Save" \
	    -variable _mmgraph(SFOB_autosave) \
	    -relief raise -bd 2
    pack $frame.autosave -side top -fill x -expand 1
    set _mmgraph(SFOB_saveall) $_mmgraph(saveall)
    frame $frame.saveselect -relief raise -bd 2
    label $frame.saveselect.topic -text "Save:" -relief flat
    pack $frame.saveselect.topic -side left -anchor nw
    frame $frame.saveselect.rb -bd 2 -relief sunken
    radiobutton $frame.saveselect.rb.saveall -text "All Data" \
	    -variable _mmgraph(SFOB_saveall) -value 1
    radiobutton $frame.saveselect.rb.savesome -text "Selected\nData" \
	    -variable _mmgraph(SFOB_saveall) -value 0
    pack $frame.saveselect.rb.saveall $frame.saveselect.rb.savesome \
	    -side top -anchor w
    pack $frame.saveselect.rb -side left
    pack $frame.saveselect -side top
}

proc SaveFileCallback { widget } {
    global _mmgraph
    set _mmgraph(saveall) $_mmgraph(SFOB_saveall)
    SaveFile [$widget GetFilename] $_mmgraph(SFOB_autosave)
    return {}
}

# Open file routines
proc OpenFile { filename } {
    global oommf _mmgpsc graph
    if {[catch {open $filename "r"} fileid]} {
	Ow_NonFatalError \
		"[Oc_Main GetAppName]: Unable to open $filename for input"
	return
    }

    # Process lines.  This can probably be streamlined considerably
    # if we ever have the need to read in large files.
    set column_list {}
    set unit_list {}
    set new_columns 1
    set ws "\[ \t\]"
    while {[gets $fileid line]>=0} {
	if {[regexp -- "^$ws*\$" $line]} {
	    continue ;# Skip blank lines
	}
	if {![string match "#*" $line]} {
	    # Data record line
	    if {$new_columns} {
		SetupDataRecordColumns $column_list $unit_list
		set new_columns 0
	    }
	    AddDataRecord $line 0
	} else {
	    # Comment or header line
	    if {[regexp -nocase -- "^#$ws*columns$ws*:$ws*\(.*\)" \
		    $line dummy value]} {
		# Column header line
		# Check for proper list structure
		if {[catch {lindex $value end}]} {
		    Ow_NonFatalError "[Oc_Main GetAppName]: Skipping illegal\
			    \"Columns\" header line in file $filename."
		} else {
		    set column_list $value
		    set new_columns 1
		}
	    } elseif {[regexp -nocase -- "^#$ws*units$ws*:$ws*\(.*\)" \
		    $line dummy value]} {
		# Units header line
		# Check for proper list structure
		if {[catch {lindex $value end}]} {
		    Ow_NonFatalError "[Oc_Main GetAppName]: Skipping illegal\
			    \"Units\" header line in file $filename."
		} else {
		    set unit_list $value
		}
	    } elseif {[regexp -nocase -- "^#$ws*title$ws*:$ws*\(.*\)" \
		    $line dummy value]} {
		# Title header line
		set _mmgpsc(title) $value
		$graph Configure -title \
			[subst -nocommands -novariables $value]
		$graph DrawAxes
	    }
	    # Currently just ignore all other header lines
	}
    }
    UpdateGraphDisplay
    close $fileid
}

proc OpenFileCallback { widget } {
    OpenFile [$widget GetFilename]
    return {}
}

proc DialogCallback { widget actionid args } {
    # Re-enable menubutton when the dialog widget is destroyed
    if {[string match DELETE $actionid]} {
        eval [join $args]
	return
    }

    # Call appropriate action proc
    set errmsg "ERROR (proc DialogCallback): Invalid actionid: $actionid"
    if {[string match SAVE $actionid]} {
	set errmsg [SaveFileCallback $widget]
    } elseif {[string match OPEN $actionid]} {
	set errmsg [OpenFileCallback $widget]
    }

    return $errmsg
}

array set graphprn {
    orient   "landscape"
    paper    "letter"
    hpos     "center"
    vpos     "center"
    units    "in"
    tmargin  "1.0"
    lmargin  "1.0"
    pwidth   "6.0"
    pheight  "6.0"
    crop     "0"
    aspect_crop {}
    aspect_nocrop "1"
}

proc PrintCallback { printwidget arrname } {
    upvar $arrname dialogprn
    global graph graphprn
    if {[info exists dialogprn]} {
	array set graphprn [array get dialogprn]
    }
    Ow_PrintDlg Print $printwidget [$graph GetCanvas] graphprn
}

proc LaunchDialog { menuname itemlabel action } {
    if { [string match "disabled" [$menuname entrycget $itemlabel -state]] } {
	return   ;# Item not enabled
    }
    global oommf
    if {[string match SAVE $action]} {
	Ow_FileDlg New dialogbox -callback DialogCallback \
		-dialog_title "[Oc_Main GetAppName]:Save File" \
		-selection_title "Append ODT Output To..." \
		-select_action_id "SAVE" \
		-filter "*.odt" \
		-optionbox_setup SaveFileOptionBoxSetup \
		-delete_callback_arg \
		  [list $menuname entryconfigure $itemlabel -state normal]
        $menuname entryconfigure $itemlabel -state disabled
    } elseif {[string match OPEN $action]} {
	Ow_FileDlg New dialogbox -callback DialogCallback \
		-dialog_title "[Oc_Main GetAppName]:Open File" \
		-selection_title "Open ODT File..." \
		-select_action_id "OPEN" \
		-filter "*.odt" \
		-delete_callback_arg \
		[list $menuname entryconfigure $itemlabel -state normal]
        $menuname entryconfigure $itemlabel -state disabled
    } elseif {[string match PRINT $action]} {
        global graphprn
	Ow_PrintDlg New dialogbox \
                -apply_callback PrintCallback \
		-dialog_title "[Oc_Main GetAppName]:Print" \
                -menu [list $menuname $itemlabel] \
                -crop_option 0 \
                -import_arrname graphprn
    } else {
	Ow_NonFatalError "Invalid action code: $action" \
		"Error in LaunchDialog proc"
	return
    }

    # Set icon
    Ow_SetIcon [$dialogbox Cget -winpath]
}

# Menu bar
set menubar .mb
foreach {filemenu xmenu ymenu optionsmenu helpmenu} \
	[Ow_MakeMenubar . $menubar File X-Axis Y-Axis Options Help] {}

$filemenu add command -label "Open..." -underline 0 \
	-command { LaunchDialog $filemenu "Open..." "OPEN" }

$filemenu add command -label "Save As..." -underline 5 \
	-command { LaunchDialog $filemenu "Save As..." "SAVE" }
$filemenu add command -label "Print..." -underline 0 \
	-command { LaunchDialog $filemenu "Print..." "PRINT" }
$filemenu add separator
$filemenu add command -label "Exit" -command { Die . } -underline 1

$xmenu configure -tearoff 1
$ymenu configure -tearoff 1
Ow_StdHelpMenu $helpmenu

proc SetPlotConfiguration { pscvar }  {
    global _mmgpsc graph
    upvar $pscvar psc
    array set _mmgpsc [array get psc]
    # Array elements: title,
    #                 autolabel,xlabel,ylabel,
    #                 xdefaultlabel,ydefaultlabel,
    #                 autolimits,xmin,xmax,ymin,ymax,
    #                 ptbufsize
    $graph Configure -title \
	    [subst -nocommands -novariables $_mmgpsc(title)]
    if {$_mmgpsc(autolabel)} {
	foreach {_mmgpsc(xlabel) _mmgpsc(ylabel)} \
		[list $_mmgpsc(xdefaultlabel) $_mmgpsc(ydefaultlabel)] {}
    }
    $graph Configure -xlabel \
	    [subst -nocommands -novariables $_mmgpsc(xlabel)] \
	    -ylabel [subst -nocommands -novariables $_mmgpsc(ylabel)]
    if {$_mmgpsc(autolimits)} {
	$graph SetGraphLimits
	foreach {_mmgpsc(xmin) _mmgpsc(ymin) _mmgpsc(xmax) _mmgpsc(ymax)} \
		[$graph GetGraphLimits] {}
    } else {
        $graph SetGraphLimits $_mmgpsc(xmin) $_mmgpsc(ymin) \
                $_mmgpsc(xmax) $_mmgpsc(ymax)
    }
    ResetBufferSize
    $graph RefreshDisplay
}
$optionsmenu add command -label "Configure..." -underline 0 \
        -command { PlotConfigure New _ \
        	  -menu "$optionsmenu {Configure...}" \
                  -import_arrname _mmgpsc \
                  -apply_callback SetPlotConfiguration
        }
$optionsmenu add command -label "clear Data" -underline 6 \
	-command {ClearDataInteractive}
$optionsmenu add command -label "Stop autosave" -underline 0 \
	-command {StopAutoSaveInteractive} -state disabled
proc MenuOptionsSaveAutoSaveState { var elt op } {
    global optionsmenu _mmgraph
    if {$_mmgraph(autosave)} {
        catch {$optionsmenu entryconfigure "Stop autosave" -state normal}
    } else {
        catch {$optionsmenu entryconfigure "Stop autosave" -state disabled}
    }
}
trace variable _mmgraph(autosave) w MenuOptionsSaveAutoSaveState

$optionsmenu add command -label "Rescale" -underline 0 \
	-command {$graph SetGraphLimits ; $graph RefreshDisplay}
set showkey 1
$optionsmenu add checkbutton -label "Key" -underline 0 \
		-variable showkey -offvalue 0 -onvalue 1 \
		-command {$graph SetKeyState $showkey}


proc PtBufOverflowCheck {} {
    global _mmgpsc data_value
    if {$_mmgpsc(ptbufsize)<1} { return } ;# No limit
    set bufstart $data_value(count,start)
    set bufend $data_value(count,end)
    set bufoverflow [expr $bufend-$bufstart-$_mmgpsc(ptbufsize)]
    if {$bufoverflow<1} { return }  ;# No overflow
    if {$bufoverflow==1} {
	# Normal overflow case (single record)
	unset data_value($bufstart)
	incr data_value(count,start)
    } else {
	# Large overflow
	set newbufstart [expr $bufstart+$bufoverflow]
	for {set i $bufstart} {$i<$newbufstart} {incr i} {
	    unset data_value($i)
	}
	set data_value(count,start) $newbufstart
    }
}

proc ResetBufferSize {} {
    global _mmgpsc graph
    set newsize $_mmgpsc(ptbufsize)
    if {[catch {expr round($newsize)} newsize] || $newsize<1} {
	set newsize 0
    }
    set _mmgpsc(ptbufsize) $newsize
    if {$newsize == $_mmgpsc(oldptbufsize)} { return }  ;# No change
    set _mmgpsc(oldptbufsize) $newsize
    PtBufOverflowCheck
    $graph Configure -ptlimit $newsize
    UpdateGraphDisplay
}

# Proc to add new labels to the x & y axis menus.  It is not currently
# possible to remove a label once added.  A global array menu_select
# stores the state of the axis menu select boxes:
#    menu_select(x)        = name of selected label, or {} if none
#    menu_select(y,$label) = 1 if box associated with label is selected,
#                            0 otherwise.
set menu_select(x) {}
proc AppendAxesSelect { new_labels } {
    global xmenu ymenu menu_select
    # Add new entries
    foreach label $new_labels {
	set menu_select(y,$label) 0
	regsub -all "\[ \r\t\n\]" $label { } compressed_label
	$xmenu add radiobutton -label $compressed_label \
		-variable menu_select(x) -value $label  \
		-command CreateGraph
	$ymenu add checkbutton -label $compressed_label \
		-variable menu_select(y,$label) \
		-offvalue 0 -onvalue 1 \
		-command "ChangeCurve \{$label\} 1"
    }
}

source [file join [file dirname [info script]] dialog.tcl]

# Graph
set graph_frame [frame .gf -borderwidth 0]
Ow_GraphWin New graph $graph_frame -default_width 340
$graph Configure -showkey $showkey
pack [$graph Cget -winpath] -side top -fill both -expand 1
pack $graph_frame -side top -fill both -expand 1
Oc_EventHandler New _ $graph Delete { Die }
bind [$graph GetCanvas] <Configure> {+
    set mywidth [%W cget -width]
    set myheight [%W cget -height]
    if {$myheight>0} {
       set graphprn(aspect_nocrop) \
	       [expr double($mywidth)/double($myheight)]
    }
}

# Proc to reset graph limits, and redraw graph
proc UpdateGraphDisplay {} {
    global graph _mmgpsc

    if {$_mmgpsc(autolimits)} {
	set limits_change [$graph SetGraphLimits]
	foreach {_mmgpsc(xmin) _mmgpsc(ymin) _mmgpsc(xmax) _mmgpsc(ymax)} \
		[$graph GetGraphLimits] {}
    } else {
        set limits_change [$graph SetGraphLimits \
		$_mmgpsc(xmin) $_mmgpsc(ymin) \
		$_mmgpsc(xmax) $_mmgpsc(ymax)]
    }
    if {$limits_change} {
	$graph RefreshDisplay
    } else {
	$graph DrawCurves
    }
}

# Proc to update graph with latest data input triples
proc UpdateGraphData { index redraw } {
    global menu_select data_value data_index graph

    if {[catch {set data_value($index)} record]} { return } ;# Out-of-bounds

    set abscissa $menu_select(x)
    if {[string match {} $abscissa]} {return} ;# No abscissa selected
    set xindex $data_index($abscissa)
    set xvalue [lindex $record $xindex]
    if {[string match {} $xvalue]} {return} ;# No abscissa value at this index

    set change_limits 0
    foreach { gxmin gymin gxmax gymax } [$graph GetGraphLimits] {}
    foreach ordinate $data_index(all) {
	if {!$menu_select(y,$ordinate)} { continue }  ;# Skip curve
	set yindex $data_index($ordinate)
	if {![string match {} [set yvalue [lindex $record $yindex]]]} {
	    foreach { xmin ymin xmax ymax } \
		    [$graph AddDataPoints $ordinate $xvalue $yvalue] {}
	    if {$xmin<$gxmin || $ymin<$gymin || \
		    $xmax>$gxmax || $ymax>$gymax} {
		set change_limits 1
	    }
	}
    }
    if {$redraw} {
	if {$change_limits} {
	    UpdateGraphDisplay
	} else {
	    $graph DrawCurves
	}
    }
    return
}

# Proc to add or delete a curve from the graph, as determined by the
# value of menu_select(y,$ordinate) (1=>display, 0=>don't display).
proc ChangeCurve { ordinate redraw } {
    global menu_select data_value data_index data_unit graph _mmgpsc
    $graph DeleteCurve $ordinate ;# Remove any previously curve by this name

    if {![info exists menu_select(y,$ordinate)]} {
	return  ;# Safety
    }
    # Count number of active curves
    set curve_count 0
    foreach curve $data_index(all) {
	if {$menu_select(y,$curve)} {
	    incr curve_count
	}
    }
    if {!$menu_select(y,$ordinate)} {
	# Curve turned off.  If this is last curve, delete y-label.
	if {$curve_count==0} {
	    if {$_mmgpsc(autolabel)} {
		set _mmgpsc(ylabel) {}
		$graph Configure -ylabel $_mmgpsc(ylabel)
		$graph DrawAxes
	    }
	    set _mmgpsc(ydefaultlabel) {}
	}
	return
    }
    # Otherwise, turn curve on...
    if [string match {} $menu_select(x)] { return 0 } ;# No abscissa selected
    set abscissa $menu_select(x)
    # If y-label is not yet set, build one from current curve.
    if {$curve_count==1} {
	set yunit $data_unit($ordinate)
	if {[string match {} $yunit]} { set yunit "Non Dim" }
	set _mmgpsc(ydefaultlabel) $yunit
	if {$_mmgpsc(autolabel)} { set _mmgpsc(ylabel) $yunit }
        $graph Configure -ylabel \
		[subst -nocommands -novariables $_mmgpsc(ylabel)]
    }
    # Write curve
    set data_list {}
    $graph NewCurve $ordinate
    set xindex $data_index($abscissa)
    set yindex $data_index($ordinate)
    for {set i $data_value(count,start)} {$i<$data_value(count,end)} {incr i} {
	set record $data_value($i)
	if {[string match {} [set xvalue [lindex $record $xindex]]] || \
		[string match {} [set yvalue [lindex $record $yindex]]]} {
	    continue
	}
	append data_list " $xvalue $yvalue"
    }
    if {![string match {} $data_list]} {
	eval {$graph AddDataPoints $ordinate} $data_list
    }
    if {$redraw} {
	UpdateGraphDisplay
    }

    return
}

# Proc to create a complete new graph, erasing any previously existing
# graph.  Return value is the number of curves in new graph.
proc CreateGraph {} {
    global menu_select data_value data_index data_unit graph _mmgpsc showkey
    set key_pos [$graph GetKeyPosition]
    $graph SetKeyState 0 ;# Temporarily hide key, if not already hidden
    $graph Reset    ;# Remove any previously existing graph
    $graph Configure -title \
	    [subst -nocommands -novariables $_mmgpsc(title)]
    if [string match {} $menu_select(x)] { return 0 } ;# No abscissa selected
    set abscissa $menu_select(x)
    set xunit " ($data_unit($abscissa))"
    if {[string match " ()" $xunit]} { set xunit {} }
    set _mmgpsc(xdefaultlabel) "$abscissa$xunit"
    if {$_mmgpsc(autolabel)} {
	set _mmgpsc(xlabel) "$abscissa$xunit"
    }
    $graph Configure -xlabel \
	    [subst -nocommands -novariables $_mmgpsc(xlabel)] \
	    -ylabel [subst -nocommands -novariables $_mmgpsc(ylabel)]
    set curve_count 0
    foreach ordinate $data_index(all) {
	if {![info exists menu_select(y,$ordinate)]} { continue } ;# Safety
	if {! $menu_select(y,$ordinate) } { continue }        ;# Skip curve
	ChangeCurve $ordinate 0
	incr curve_count
    }
    $graph DrawAxes
    UpdateGraphDisplay

    $graph SetKeyState $showkey $key_pos
    ## Reset original key display state

    return $curve_count
}

# Data storage
#   Data is stored in two global arrays: data_value and data_index.
# The first is an array of lists indexed by import order, e.g.,
#                data_value(4)
# Each element is a list of values.  The data_index array provides a
# mapping from label names to position inside the data_value lists.
# This position is constant for each label across all data_value entries.
# A data_value list entry is {} if no data field corresponding to that
# entry was entered on that import record.  Also, some data_value() lists
# may be longer than others.  If data_index(label) points past the end
# of a data_value() list, then the corresponding value is {} i.e.,
# missing.
#   The data_value array also contains two elements,
# data_value(count,start) and data_value(count,end) which denote the first
# and last data_value record indices stored in the array, i.e.,
#      data_value(count,start) <= i < data_value(count,end)
# Similarly, the data_index(count) element contains the length of the
# longest data_value() list.  data_index(all) is an ordered list of
# the data labels.
#   There is currently no way to remove a label once it has been added.
#   A previous version of this code stored each data item in an array
# element (as opposed to grouping each record together in a list).  For
# 10 element data records, data import appears to be about 20% faster the
# old way, but storage costs are about 4 times higher (about 1K per record
# as opposed to about 250 bytes per record).  Moreover, it is probably
# possible to streamline the new code somewhat.
#
#   The return value for the import StoreDataTable routine is a list of all
# new labels (if any).  Each call to "StoreDataTable" represents 1 record.
# If a labels occurs more than once in an import record, only the last
# instance is stored.
#
# Addenda, 21-May-1998 mjd: As the first step towards a move away from
#  importing triples towards separate column and bare data records, I
#  have added
#     data_use()      : Array mapping storage list index positions to
#                     : positions in the incoming bare data record lists.
#     proc SetupDataRecordColumns { columns_list unit_list }
#                     : Adds column labels to data_index and data_unit
#                     : lists as necessary, and sets up data_index(use)
#                     : for subsequent use by the AddDataRecord proc.
#     proc AddDataRecord { data_list redraw }
#                     : Adds data to the data_value array, via the
#                     : data_index(use) ordering.

set data_value(count,start) 0
set data_value(count,end) 0
set data_index(count) 0   ;# Current maximum length of data_value lists.
set data_index(all) {}
set data_index(use) {}
proc StoreDataTable { triples } {
    global data_value data_index data_unit _mmgraph
    set new_labels {}
    foreach trip $triples {
	foreach {label unit value} $trip {}
	if {[catch {set data_index($label)} index]} {
	    # New label
	    set index $data_index(count)
	    set data_index($label) $index
	    lappend data_index(all) $label
	    set data_unit($label) $unit
	    incr data_index(count)
	    lappend new_labels $label
	}
	set new_value($index) $value    ;# Store temporarily in an array
    }
    # Construct data list
    set new_list {}
    for {set index 0} {$index<$data_index(count)} {incr index} {
	if {[catch {lappend new_list $new_value($index)}]} {
	    lappend new_list {}      ;# Missing element
	}
    }
    set recno $data_value(count,end) ;  incr data_value(count,end)
    set data_value($recno) $new_list
    if {$_mmgraph(autosave)} { SaveRecord $new_list }
    PtBufOverflowCheck
    if {![string match {} $new_labels]} {
	AppendAxesSelect $new_labels
    }
    UpdateGraphData $recno 1
    return $new_labels
}

proc SetupDataRecordColumns { columns_list unit_list } {
    # Initialize column input correspondance for bare
    # data record input.
    global data_index data_unit data_use
    catch {unset data_use}
    set new_labels {}
    set offset 0
    foreach label $columns_list unit $unit_list {
	if {[catch {set data_index($label)} index]} {
	    # New column
	    set index $data_index(count)
	    set data_index($label) $index
	    lappend data_index(all) $label
	    set data_unit($label) $unit
	    incr data_index(count)
	    lappend new_labels $label
	}
	set data_use($index) $offset
	incr offset
    }
    if {![string match {} $new_labels]} {
	AppendAxesSelect $new_labels
    }
    return $new_labels
}

proc AddDataRecord { data_list redraw } {
    # Add bare data record, ordered by the data_use() array.
    global data_index data_value data_use _mmgraph

    # Construct data list
    set new_list {}
    for {set index 0} {$index<$data_index(count)} {incr index} {
	if {[catch {set offset $data_use($index)}]} {
	    lappend new_list {}      ;# Missing element
	} else {
	    lappend new_list [lindex $data_list $offset]
	}
    }
    set recno $data_value(count,end) ;  incr data_value(count,end)
    set data_value($recno) $new_list
    if {$_mmgraph(autosave)} { SaveRecord $new_list } ;# Save to disk
    PtBufOverflowCheck
    UpdateGraphData $recno $redraw
}

proc ClearData {} {
    global data_value graph
    unset data_value
    set data_value(count,start) 0
    set data_value(count,end) 0
    $graph DeleteAllData
}

proc ClearDataInteractive {} {
    set response [Ow_Dialog 1 "[Oc_Main GetAppName]: Clear Data" warning \
	    "Are you *sure* want to erase all data in stored\
	    in this graph widget?" \
	    {} 1 "Yes" "No"]
    if { $response == 0 } { ClearData }
}

proc StopAutoSaveInteractive {} {
    set response [Ow_Dialog 1 "[Oc_Main GetAppName]: Stop Auto Save" warning \
	    "Are you *sure* want to stop automatically saving data received\
	    by this graph widget?" \
	    {} 1 "Yes" "No"]
    if { $response == 0 } { CloseAutoSave }
}

if {!$no_net} {
    # Protocol
    package require Net
    Net_Protocol New protocol -name "OOMMF DataTable protocol 0.0"
    $protocol AddMessage start HasGui {} {
	return [list start [list 0 0]]
    }
    $protocol AddMessage start DataTable { triples } {
	StoreDataTable $triples
	return [list start [list 0 1]]
    }

    Net_Server New server -protocol $protocol -alias [Oc_Main GetAppName] \
            -log Oc_Nop
    $server Start 0
} else {
    proc AddTriples { triples } {
	StoreDataTable $triples
    }
}

wm title . [Oc_Main GetAppName]<[pid]>
Ow_SetIcon .

# Enter event loop
if {!$no_event_loop} {
    vwait forever
}
